__all__ = ('escape_codes', 'parse_colors')

def esc(*x):
    return f"\033[{';'.join(x)}m"


escape_codes = {
    'reset': esc('0'),
    'bold': esc('01'),
    'thin': esc('02')
}

COLORS = [
    'black',
    'red',
    'green',
    'yellow',
    'blue',
    'purple',
    'cyan',
    'white'
]

PREFIXES = [
    ('3', ''), ('01;3', 'bold_'), ('02;3', 'thin_'),
    ('3', 'fg_'), ('01;3', 'fg_bold_'), ('02;3', 'fg_thin_'),
    ('4', 'bg_'), ('10', 'bg_bold_')
]

for prefix, prefix_name in PREFIXES:
    for code, name in enumerate(COLORS):
        escape_codes[prefix_name + name] = esc(prefix + str(code))


def parse_colors(name):
    return escape_codes[name]


from dsb_spider.log import getLogger
from dsb_spider.log.ex import DsbException
logger = getLogger('dsb')
from dsb_spider.utils import hash_args

_TASK_FUNC_REGISTERS = {}

class _FuncRegistry():
    def __init__(self, name:str):
        self.name = name
        self.funcs = {}
        
    def update(self, funcs_iter):
        if hasattr(funcs_iter, 'items'):
            funcs_iter = funcs_iter.items()
        for key, value in funcs_iter:
            self[key] = value
            
    def __setitem__(self, key, value):
        if not callable(value):
            raise DsbException(f'func {key} register failed')
        if not key:
            raise ValueError('func name must not be none')
        self.funcs[key] = value
        
    def __getitem__(self, item):
        return self.funcs[item]
    
    def __delitem__(self, key):
        del self.funcs[key]
        
    def __iter__(self):
        return iter(self.funcs)
    
    def items(self):
        return list(self.funcs.items())
    
    def iteritems(self):
        return iter(self.funcs.items())
    
    def clear(self):
        self.funcs.clear()
        
    def __call__(self, obj):
        self[obj.__name__] = obj
        return obj
    
    def __deco(self, name, obj):
        self[name] = obj
        return obj
    
    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            return
    
    def __repr__(self):
        return f'<{self.__class__.__name__}|{self.name}>'
    
    def __str__(self):
        return f'<{self.__class__.__name__}|{self.name}>'

def getTaskFuncRegister(name: str):
    try:
        return _TASK_FUNC_REGISTERS[name]
    except KeyError:
        register = _TASK_FUNC_REGISTERS[name] = _FuncRegistry(name)
        return register

# 使用单例保证相同任务运行的唯一性
class TaskSingleton(type):
    _instances = {}
    
    def __call__(cls,*args, **kwargs):
        key = hash_args(*args, **kwargs)
        try:
            instance = cls._instances[key]
        except KeyError:
            instance = cls._instances[key] = super().__call__(*args, **kwargs)
            instance.hash_key = key
        return instance
    
    @classmethod
    def clear(cls, key):
        try:
            del cls._instances[key]
        except KeyError:
            pass

# 定义任务状态: stoped -> ready -> stoped , 只有ready状态可run
# ready:

class StopedState():
    @staticmethod
    def ready(task):
        task.new_state(ReadyState)
        task.do_ready()
    
    @staticmethod
    def run(task):
        raise RuntimeError('Not ready')
    
    @staticmethod
    def stop(task):
        raise RuntimeError('Already stoped')

class ReadyState():
    @staticmethod
    def ready(task):
        raise RuntimeError('Already ready')

    @staticmethod
    def run(task):
        task.do_task()

    @staticmethod
    def stop(task):
        task.new_state(StopedState)
        task.do_stop()
    
# 定义任务阶段： ready -> run -> finish

class Task(object, metaclass=TaskSingleton):
    def __init__(self,name, listener=None, **kwargs):
        self.task_name = name
        self.listenter=listener
        self.register_task_funcs()
        self.new_state(StopedState)
        if hasattr(self, 'init_infos'):
            self.init_infos(**kwargs)
        
    def register_task_funcs(self):
        registry = _TASK_FUNC_REGISTERS[self.task_name]
        for key, value in registry.items():
            setattr(self, key, value)
    
    def new_state(self, state):
        self._state = state
    
    def ready(self):
        self._state.ready(self)
    
    def run(self):
        self._state.run(self)
    
    def finish(self):
        if hasattr(self, 'hash_key'):
            TaskSingleton.clear(self.hash_key)
        self._state.stop(self)
    
if __name__ == '__main__':
    haha = getTaskFuncRegister('haha')
    @haha
    def do_task():
        print('go')

    task = Task('haha')
    task.ready()
    task.run()
    task.finish()





'''
所有exception，都要继承DsbException，并注明ex_code以作区分
'''


class DsbException(Exception):
    ex_code = 1
    ex_msg = 'error~~'
    
    def __init__(self, ex_msg=None):
        self.ex_msg = ex_msg or self.ex_msg
        super().__init__(ex_msg)
    
    def __str__(self):
        return self.ex_msg


class InvalidProduct(DsbException):
    ex_code = 2
    ex_msg = 'Product Not Found'


class EndTask(DsbException):
    ex_code = 3
    
    def __init__(self, msg=''):
        msg = 'end this task. {}'.format(msg)
        super().__init__(msg)


class CustomMadeProduct(DsbException):
    ex_code = 4
    ex_msg = 'it is a custom-made product, which we can not sell now'


class OutOfStock(DsbException):
    ex_code = 5
    ex_msg = 'Out Of Stock'


class TemporarilyOutOfStock(DsbException):
    ex_code = 6
    ex_msg = 'TemporarilyOutOfStock'


class EmptyResponse(DsbException):
    ex_code = 7
    ex_msg = 'EmptyResponse, please try again later.'


class NeedProxy(DsbException):
    ex_code = 8
    ex_msg = 'Access denied,need proxy'


class NeedProxyPool(DsbException):
    ex_code = 9
    ex_msg = 'Connection error,proxy pool need run'


class NotGood(DsbException):
    ex_code = 10
    ex_msg = 'Something wrong'


class CategoryNotFound(DsbException):
    ex_code = 11
    ex_msg = 'category not found'


class NeedLocalInsert(DsbException):
    ex_code = 12
    ex_msg = '这类商品需要本地手动入库'


class RedirectError(DsbException):
    ex_code = 13
    ex_msg = 'redirect error'


class NeedVPN(DsbException):
    ex_code = 14
    ex_msg = 'Need VPN!!'


class InvalidCategory(DsbException):
    ex_code = 15
    ex_msg = '违禁类目'


class CanNotFindImage(DsbException):
    ex_code = 16
    ex_msg = '图片未找到'


class CanNotFindSize(DsbException):
    ex_code = 17
    ex_msg = 'size未找到'


class AccessDenied(DsbException):
    ex_code = 18
    ex_msg = 'ip被封'


class TooManyRequestRetries(DsbException):
    ex_code = 19
    
    def __init__(self, msg=''):
        msg = '请求重试过多: {}'.format(msg)
        super().__init__(msg)


class OnceProduct(DsbException):
    ex_code = 20
    ex_msg = '这种商品只做一次入库，不需要再入了'


class CrmError(DsbException):
    ex_code = 21
    ex_msg = '从CRM 取任务 出了问题?'